import simplejson as json
from django.conf import settings as django_settings
from django.http import HttpResponse
from cio.conf import settings
import djedi

# TODO: Switch simplejson to ujson or other?


class JSONResponseMixin(object):
    """
    A mixin that can be used to render a JSON response.
    """
    response_class = HttpResponse

    def render_to_json(self, context, **response_kwargs):
        """
        Returns a JSON response, transforming 'context' to make the payload.
        """
        response_kwargs['content_type'] = 'application/json'
        return self.response_class(
            self.convert_context_to_json(context),
            **response_kwargs
        )

    def convert_context_to_json(self, context):
        """Convert the context dictionary into a JSON object"""
        return json.dumps(context, indent=4, for_json=True)


class DjediContextMixin(object):

    def get_context_data(self, **context):
        theme = settings.THEME

        if '/' not in theme:
            theme = '{static}djedi/themes/{theme}/theme.css'.format(static=django_settings.STATIC_URL, theme=theme)

        context['THEME'] = theme
        context['VERSION'] = djedi.__version__

        return context



import django
from django.contrib import admin
from django.shortcuts import render_to_response

from ..compat import include, patterns, url

admin.autodiscover()

if django.VERSION < (2, 0):
    admin_urls = include(admin.site.urls)
else:
    admin_urls = admin.site.urls

urlpatterns = patterns(
    url(r'^$', lambda r: render_to_response('index.html'), name='index'),
    url(r'^adm1n/', admin_urls),
    url(r'^djed1/', include('djedi.urls', namespace='admin'))
)



import logging

from django.db import IntegrityError
from cio.backends.base import DatabaseBackend
from cio.backends.exceptions import NodeDoesNotExist, PersistenceError

# Use absolute import here or Django 1.7 complains about duplicate models.
from djedi.backends.django.db.models import Node

logger = logging.getLogger(__name__)


class DjangoModelStorageBackend(DatabaseBackend):

    scheme = 'db'

    def __init__(self, **config):
        super(DjangoModelStorageBackend, self).__init__(**config)

    def get_many(self, uris):
        storage_keys = dict((self._build_key(uri), uri) for uri in uris)
        stored_nodes = Node.objects.filter(key__in=storage_keys.keys())
        stored_nodes = stored_nodes.values_list('key', 'content', 'plugin', 'version', 'is_published', 'meta')

        # Filter matching nodes
        nodes = {}
        for key, content, plugin, version, is_published, meta in stored_nodes:
            uri = storage_keys[key]

            # Assert requested plugin matches
            if uri.ext in (None, plugin):

                # Assert version matches or node is published
                if (uri.version == version) or (is_published and not uri.version):
                    meta = self._decode_meta(meta, is_published=is_published)
                    nodes[uri] = {
                        'uri': uri.clone(ext=plugin, version=version),
                        'content': content,
                        'meta': meta
                    }

        return nodes

    def publish(self, uri, **meta):
        node = self._get(uri)

        if not node.is_published:
            # Assign version number
            if not node.version.isdigit():
                revisions = Node.objects.filter(key=node.key).values_list('version', flat=True)
                version = self._get_next_version(revisions)
                node.version = version

            # Un publish any other revision
            Node.objects.filter(key=node.key).update(is_published=False)

            # Publish this version
            self._update_meta(node, meta)
            node.is_published = True
            node.save()

        return self._serialize(uri, node)

    def get_revisions(self, uri):
        key = self._build_key(uri)
        nodes = Node.objects.filter(key=key).order_by('date_created')
        revisions = nodes.values_list('plugin', 'version', 'is_published')
        return [(key.clone(ext=plugin, version=version), is_published) for plugin, version, is_published in revisions]

    def _get(self, uri):
        key = self._build_key(uri)
        nodes = Node.objects.filter(key=key)

        if uri.ext:
            nodes = nodes.filter(plugin=uri.ext)
        if uri.version:
            nodes = nodes.filter(version=uri.version)
        else:
            nodes = nodes.filter(is_published=True)

        try:
            return nodes.get()
        except Node.DoesNotExist:
            raise NodeDoesNotExist('Node for uri "%s" does not exist' % uri)

    def _create(self, uri, content, **meta):
        try:
            meta = self._encode_meta(meta)
            return Node.objects.create(
                key=self._build_key(uri),
                content=content,
                plugin=uri.ext,
                version=uri.version,
                is_published=False,
                meta=meta
            )
        except IntegrityError as e:
            raise PersistenceError('Failed to create node for uri "%s"; %s' % (uri, e))

    def _update(self, uri, content, **meta):
        node = self._get(uri)

        self._update_meta(node, meta)

        node.content = content
        node.plugin = uri.ext
        node.version = uri.version
        node.save()

        return node

    def _delete(self, node):
        node.delete()

    def _serialize(self, uri, node):
        meta = self._decode_meta(node.meta, is_published=node.is_published)
        return {
            'uri': uri.clone(ext=node.plugin, version=node.version),
            'content': node.content,
            'meta': meta
        }

    def _update_meta(self, node, meta):
        node.meta = self._merge_meta(node.meta, meta)



import os

import simplejson as json
from django.core.files import File
from django.test import Client
from django.utils.http import urlquote

import cio
import cio.conf
from cio.backends import storage
from cio.backends.exceptions import NodeDoesNotExist, PersistenceError
from cio.plugins import plugins
from cio.utils.uri import URI
from djedi.tests.base import ClientTest, DjediTest, UserMixin
from djedi.utils.encoding import smart_unicode

from ..compat import reverse


def json_node(response, simple=True):
    node = json.loads(response.content)
    if simple and 'meta' in node:
        del node['meta']
    return node


class PermissionTest(DjediTest, UserMixin):

    def setUp(self):
        super(PermissionTest, self).setUp()
        self.master = self.create_djedi_master()
        self.apprentice = self.create_djedi_apprentice()

    def test_permissions(self):
        client = Client()
        url = reverse('admin:djedi:api', args=['i18n://sv-se@page/title'])

        response = client.get(url)
        self.assertEqual(response.status_code, 403)

        logged_in = client.login(username=self.master.username, password='test')
        self.assertTrue(logged_in)
        response = client.get(url)
        self.assertEqual(response.status_code, 404)

        client.logout()
        logged_in = client.login(username=self.apprentice.username, password='test')
        self.assertTrue(logged_in)
        response = client.get(url)
        self.assertEqual(response.status_code, 404)


class PrivateRestTest(ClientTest):

    def get_api_url(self, url_name, uri):
        return reverse('admin:djedi:' + url_name, args=[urlquote(urlquote(uri, ''), '')])

    def get(self, url_name, uri):
        url = self.get_api_url(url_name, uri)
        return self.client.get(url)

    def post(self, url_name, uri, data):
        url = self.get_api_url(url_name, uri)
        return self.client.post(url, data)

    def put(self, url_name, uri, data=None):
        url = self.get_api_url(url_name, uri)
        return self.client.put(url, data=data or {})

    def delete(self, url_name, uri):
        url = self.get_api_url(url_name, uri)
        return self.client.delete(url)

    def test_get(self):
        response = self.get('api', 'i18n://sv-se@page/title')
        self.assertEqual(response.status_code, 404)

        cio.set('i18n://sv-se@page/title.md', u'# Djedi', publish=False)

        response = self.get('api', 'i18n://sv-se@page/title')
        self.assertEqual(response.status_code, 404)

        response = self.get('api', 'i18n://sv-se@page/title#draft')
        self.assertEqual(response.status_code, 200)
        node = json_node(response)
        self.assertKeys(node, 'uri', 'content')
        self.assertEqual(node['uri'], 'i18n://sv-se@page/title.md#draft')
        self.assertRenderedMarkdown(node['content'], u'# Djedi')

    def test_load(self):
        response = self.get('api.load', 'i18n://sv-se@page/title')
        self.assertEqual(response.status_code, 200)
        json_content = json.loads(response.content)
        self.assertEqual(json_content['uri'], 'i18n://sv-se@page/title.txt')
        self.assertIsNone(json_content['data'])
        self.assertEqual(len(json_content['meta'].keys()), 0)

        # TODO: Should get 404
        # response = self.get('api.load', 'i18n://sv-se@page/title#1')
        # self.assertEqual(response.status_code, 404)

        cio.set('i18n://sv-se@page/title.md', u'# Djedi')

        response = self.get('api.load', 'sv-se@page/title')
        self.assertEqual(response.status_code, 200)
        node = json_node(response, simple=False)
        meta = node.pop('meta', {})
        content = u'# Djedi' if cio.PY26 else u'<h1>Djedi</h1>'
        self.assertDictEqual(node, {'uri': 'i18n://sv-se@page/title.md#1', 'data': u'# Djedi', 'content': content})
        self.assertKeys(meta, 'modified_at', 'published_at', 'is_published')

        response = self.get('api.load', 'i18n://sv-se@page/title#1')
        json_content = json.loads(response.content)
        self.assertEqual(json_content['uri'], 'i18n://sv-se@page/title.md#1')

        self.assertEqual(len(cio.revisions('i18n://sv-se@page/title')), 1)

    def test_set(self):
        response = self.post('api', 'i18n://page/title', {'data': u'# Djedi'})
        self.assertEqual(response.status_code, 400)

        response = self.post('api', 'i18n://sv-se@page/title.txt', {'data': u'# Djedi', 'data[extra]': u'foobar'})
        self.assertEqual(response.status_code, 400)

        uri = 'i18n://sv-se@page/title.md'
        response = self.post('api', uri, {'data': u'# Djedi', 'meta[message]': u'lundberg'})
        self.assertEqual(response.status_code, 200)
        node = json_node(response, simple=False)
        meta = node.pop('meta')
        content = u'# Djedi' if cio.PY26 else u'<h1>Djedi</h1>'
        self.assertDictEqual(node, {'uri': 'i18n://sv-se@page/title.md#draft', 'content': content})
        self.assertEqual(meta['author'], u'master')
        self.assertEqual(meta['message'], u'lundberg')

        node = cio.get(uri, lazy=False)
        self.assertIsNone(node.content)
        cio.publish(uri)
        node = cio.get(uri, lazy=False)
        self.assertEqual(node.uri, 'i18n://sv-se@page/title.md#1')
        self.assertRenderedMarkdown(node.content, u'# Djedi')

        response = self.post('api', node.uri, {'data': u'# Djedi', 'meta[message]': u'Lundberg'})
        node = json_node(response, simple=False)
        self.assertEqual(node['meta']['message'], u'Lundberg')

        with self.assertRaises(PersistenceError):
            storage.backend._create(URI(node['uri']), None)

    def test_delete(self):
        response = self.delete('api', 'i18n://sv-se@page/title')
        self.assertEqual(response.status_code, 404)

        node = cio.set('i18n://sv-se@page/title.md', u'# Djedi')

        response = self.delete('api', node.uri)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(smart_unicode(response.content), u'')

        with self.assertRaises(NodeDoesNotExist):
            storage.get('i18n://sv-se@page/title')

        node = cio.get('i18n://page/title', lazy=False)
        self.assertIsNone(node.content)

    def test_publish(self):
        node = cio.set('sv-se@page/title', u'Djedi', publish=False)

        response = self.get('api', 'i18n://sv-se@page/title')
        self.assertEqual(response.status_code, 404)

        response = self.put('api.publish', node.uri)
        self.assertEqual(response.status_code, 200)

        response = self.get('api', 'i18n://sv-se@page/title')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json_node(response), {'uri': 'i18n://sv-se@page/title.txt#1', 'content': u'Djedi'})
        response = self.put('api.publish', 'i18n://sv-se@foo/bar.txt#draft')

        self.assertEqual(response.status_code, 404)

    def test_revisions(self):
        cio.set('sv-se@page/title', u'Djedi 1')
        cio.set('sv-se@page/title', u'Djedi 2')

        response = self.get('api.revisions', 'sv-se@page/title')
        self.assertEqual(response.status_code, 200)

        content = json.loads(response.content)
        self.assertEqual(content, [['i18n://sv-se@page/title.txt#1', False], ['i18n://sv-se@page/title.txt#2', True]])

    def test_render(self):
        response = self.post('api.render', 'foo', {'data': u'# Djedi'})
        assert response.status_code == 404

        response = self.post('api.render', 'md', {'data': u'# Djedi'})
        assert response.status_code == 200
        self.assertRenderedMarkdown(smart_unicode(response.content), u'# Djedi')

        response = self.post('api.render', 'img', {'data': json.dumps({
            'url': '/foo/bar.png',
            'width': '64',
            'height': '64'
        })})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(smart_unicode(response.content), u'<img alt="" height="64" src="/foo/bar.png" width="64" />')

    def test_editor(self):
        response = self.get('cms.editor', 'sv-se@page/title.foo')
        self.assertEqual(response.status_code, 404)

        response = self.get('cms.editor', 'sv-se@page/title')
        self.assertEqual(response.status_code, 404)

        for ext in plugins:
            response = self.get('cms.editor', 'sv-se@page/title.' + ext)
            self.assertEqual(response.status_code, 200)
            assert set(response.context_data.keys()) == set(('THEME', 'VERSION', 'uri',))
            self.assertNotIn(b'document.domain', response.content)

        with cio.conf.settings(XSS_DOMAIN='foobar.se'):
            response = self.post('cms.editor', 'sv-se@page/title', {'data': u'Djedi'})
            self.assertEqual(response.status_code, 200)
            self.assertIn(b'document.domain = "foobar.se"', response.content)

    def test_upload(self):
        tests_dir = os.path.dirname(os.path.abspath(__file__))
        image_path = os.path.join(tests_dir, 'assets', 'image.png')

        form = {
            'data[width]': u'64',
            'data[height]': u'64',
            'data[crop]': u'64,64,128,128',
            'data[id]': u'vw',
            'data[class]': u'year-53',
            'data[alt]': u'Zwitter',
            'meta[comment]': u'VW'
        }
        response = self.post('api', 'i18n://sv-se@header/logo.img', form)
        self.assertEqual(response.status_code, 200)

        with open(image_path, "rb") as image:
            file = File(image, name=image_path)
            form['data[file]'] = file
            response = self.post('api', 'i18n://sv-se@header/logo.img', form)
            self.assertEqual(response.status_code, 200)

            node = json_node(response, simple=False)
            meta = node.pop('meta')
            uri, content = node['uri'], node['content']
            self.assertEqual(uri, 'i18n://sv-se@header/logo.img#draft')
            self.assertEqual(meta['comment'], u'VW')
            html = u'<img ' \
                   u'alt="Zwitter" ' \
                   u'class="year-53" ' \
                   u'height="64" ' \
                   u'id="vw" ' \
                   u'src="/media/djedi/img/03/5e/5eba6fc2149822a8dbf76cd6978798f2ddc4ac34.png" ' \
                   u'width="64" />'
            self.assertEqual(content, html)

            # Post new resized version
            node = cio.load(uri)
            del form['data[file]']
            del form['data[crop]']
            form['data[width]'] = form['data[height]'] = u'32'
            form['data[filename]'] = node['data']['filename']

            response = self.post('api', 'i18n://sv-se@header/logo.img', form)
            self.assertEqual(response.status_code, 200)


class PublicRestTest(ClientTest):

    def test_api_root_not_found(self):
        url = reverse('admin:djedi:rest:api-base')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)

    def test_embed(self):
        url = reverse('admin:djedi:rest:embed')
        response = self.client.get(url)
        html = smart_unicode(response.content)

        self.assertIn(u'iframe id="djedi-cms"', html)
        cms_url = u'http://testserver' + reverse('admin:djedi:cms')
        self.assertIn(cms_url, html)
        self.assertNotIn(u'window.DJEDI_NODES', html)
        self.assertNotIn(u'document.domain', html)

        with cio.conf.settings(XSS_DOMAIN='foobar.se'):
            response = self.client.get(url)
            self.assertIn(b'document.domain = "foobar.se"', response.content)

        self.client.logout()
        response = self.client.get(url)
        self.assertEqual(response.status_code, 204)

    def test_nodes(self):
        with self.assertCache(sets=1):
            cio.set('sv-se@rest/label/email', u'E-post')

        with self.assertDB(calls=1), self.assertCache(calls=1, misses=1, hits=1, sets=0):
            url = reverse('admin:djedi:rest:nodes')
            response = self.client.post(url, json.dumps({
                'rest/page/body.md': u'# Foo Bar',
                'rest/label/email': u'E-mail',
            }), content_type='application/json')

        json_content = json.loads(response.content)

        self.assertIn('i18n://sv-se@rest/page/body.md', json_content.keys())
        self.assertEqual(json_content['i18n://sv-se@rest/page/body.md'], u'<h1>Foo Bar</h1>')
        self.assertIn('i18n://sv-se@rest/label/email.txt#1', json_content.keys())
        self.assertEqual(json_content['i18n://sv-se@rest/label/email.txt#1'], u'E-post')



from ..compat import include, patterns, url
from .api import LoadApi, NodeApi, NodeEditor, PublishApi, RenderApi, RevisionsApi
from .cms import DjediCMS

app_name = 'djedi'

urlpatterns = patterns(
    url(r'^$', DjediCMS.as_view(), name='cms'),
    url(r'^node/(?P<uri>.+)/editor$', NodeEditor.as_view(), name='cms.editor'),
    url(r'^node/(?P<uri>.+)/load$', LoadApi.as_view(), name='api.load'),
    url(r'^node/(?P<uri>.+)/publish$', PublishApi.as_view(), name='api.publish'),
    url(r'^node/(?P<uri>.+)/revisions$', RevisionsApi.as_view(), name='api.revisions'),
    url(r'^node/(?P<uri>.+)$', NodeApi.as_view(), name='api'),
    url(r'^plugin/(?P<ext>\w+)$', RenderApi.as_view(), name='api.render'),
    url(r'^api/', include('djedi.rest.urls', namespace='rest'))
)



import logging

from django import template
from django.utils.html import mark_safe

import cio.conf
from cio.pipeline import pipeline
from djedi.auth import has_permission
from djedi.utils.templates import render_embed


register = template.Library()
logger = logging.getLogger(__name__)


@register.simple_tag(takes_context=True)
def djedi_admin(context):
    output = u''

    if has_permission(context.get('request')):
        defaults = dict((node.uri.clone(version=None), node.initial) for node in pipeline.history.list('get'))
        output = render_embed(nodes=defaults)

    # Clear pipeline
    pipeline.clear()

    return output


@register.simple_tag
def djedi_xss_domain():
    domain = cio.conf.settings.get('XSS_DOMAIN')
    if domain:
        return mark_safe(u'<script>document.domain = "{domain}";</script>'.format(domain=domain))

    return u''



import shutil

import cio
import django

from contextlib import contextmanager

from django.conf import settings
from django.contrib.auth.models import User, Group
from django.test import Client
from django.test import TransactionTestCase

from cio.conf import settings as cio_settings
from djedi import configure


if django.VERSION < (1, 8):
    DEBUG_CURSOR_ATTR = 'use_debug_cursor'
else:
    DEBUG_CURSOR_ATTR = 'force_debug_cursor'


class DjediTest(TransactionTestCase):

    def setUp(self):
        from cio.environment import env
        from cio.backends import cache
        from cio.pipeline import pipeline
        from cio.plugins import plugins

        env.reset()
        cache.clear()
        pipeline.clear()
        plugins.load()

    def tearDown(self):
        shutil.rmtree(settings.MEDIA_ROOT, ignore_errors=True)

    @contextmanager
    def settings(self, *args, **kwargs):
        with super(DjediTest, self).settings(*args, **kwargs):
            with cio_settings():
                configure()
                yield


class AssertionMixin(object):

    def assertKeys(self, dict, *keys):
        self.assertEqual(set(dict.keys()), set(keys))

    @contextmanager
    def assertCache(self, calls=-1, hits=-1, misses=-1, sets=-1):
        from cio.backends import cache

        _cache = cache.backend._cache

        _cache.calls = 0
        _cache.hits = 0
        _cache.misses = 0
        _cache.sets = 0

        yield

        if calls >= 0:
            assert _cache.calls == calls, '%s != %s' % (_cache.calls, calls)
        if hits >= 0:
            assert _cache.hits == hits, '%s != %s' % (_cache.hits, hits)
        if misses >= 0:
            assert _cache.misses == misses, '%s != %s' % (_cache.misses, misses)
        if sets >= 0:
            assert _cache.sets == sets, '%s != %s' % (_cache.sets, sets)

    @contextmanager
    def assertDB(self, calls=-1, selects=-1, inserts=-1, updates=-1):
        from django.db import connection

        pre_debug_cursor = getattr(connection, DEBUG_CURSOR_ATTR)
        setattr(connection, DEBUG_CURSOR_ATTR, True)
        pre_num_queries = len(connection.queries)

        yield

        queries = connection.queries[pre_num_queries:]
        num_queries = len(queries)
        setattr(connection, DEBUG_CURSOR_ATTR, pre_debug_cursor)

        if calls >= 0:
            assert num_queries == calls, '%s != %s' % (num_queries, calls)
        if selects >= 0:
            num_selects = len([q for q in queries if q['sql'].startswith('SELECT')])
            assert num_selects == selects, '%s != %s' % (num_selects, selects)
        if inserts >= 0:
            num_inserts = len([q for q in queries if q['sql'].startswith('INSERT')])
            assert num_inserts == inserts, '%s != %s' % (num_inserts, inserts)
        if updates >= 0:
            num_updates = len([q for q in queries if q['sql'].startswith('UPDATE')])
            assert num_updates == updates, '%s != %s' % (num_updates, updates)

    def assertRenderedMarkdown(self, value, source):
        if cio.PY26:
            self.assertEqual(value, source)  # Markdown lacks Support for python 2.6
        else:
            from markdown import markdown
            rendered = markdown(source)
            self.assertEqual(value, rendered)


class UserMixin(object):

    def create_djedi_master(self):
        user = User.objects.create_superuser('master', 'master@djedi.io', 'test')
        return user

    def create_djedi_apprentice(self):
        user = User.objects.create_user('apprentice', email='apprentice@djedi.io', password='test')
        group, _ = Group.objects.get_or_create(name='Djedi')
        user.is_staff = True
        user.groups.add(group)
        user.save()
        return user


class ClientTest(DjediTest, UserMixin, AssertionMixin):

    def setUp(self):
        super(ClientTest, self).setUp()
        master = self.create_djedi_master()
        client = Client(enforce_csrf_checks=True)
        client.login(username=master.username, password='test')
        self.client = client



from django.http import Http404

from ..compat import patterns, url
from .api import EmbedApi, NodesApi

app_name = 'rest'


def not_found(*args, **kwargs):
    raise Http404


urlpatterns = patterns(
    url(r'^$', not_found, name='api-base'),
    url(r'^embed/$', EmbedApi.as_view(), name='embed'),
    url(r'^nodes/$', NodesApi.as_view(), name='nodes'),
)



import djedi
from cio.backends import storage


# Setup content-io configuration
djedi.configure()

# Setup node django model based on backend
if storage.backend.scheme == 'db':
    from .backends.django.db.models import Node
    Node



import cio
from django.contrib.auth.models import User
from django.template import TemplateSyntaxError
from cio.backends import cache
from cio.pipeline import pipeline
from djedi.templatetags.template import register
from djedi.tests.base import DjediTest, AssertionMixin
from .compat import cmpt_context, get_template_from_string


class TagTest(DjediTest, AssertionMixin):

    def render(self, source, context=None):
        source = u'{% load djedi_tags %}' + source.strip()
        context = cmpt_context(context or {})
        return get_template_from_string(source).render(context).strip()

    def test_node_tag(self):
        html = self.render(u"{% node 'page/title' edit=False %}")
        assert html == u''

        cio.set('i18n://sv-se@page/title.txt', u'Djedi')
        cio.set('i18n://sv-se@page/body.txt', u'Lightning fast!')

        with self.assertCache(calls=1, misses=0):
            with self.assertDB(calls=0):
                html = self.render(u"<h1>{% node 'page/title' edit=False %}</h1><p>{% node 'page/body' edit=False %}</p>")
                assert html == u'<h1>Djedi</h1><p>Lightning fast!</p>'

        cache.clear()

        with self.assertCache(calls=1, misses=2):
            with self.assertDB(calls=1):
                html = self.render(u"<h1>{% node 'page/title' edit=False %}</h1><p>{% node 'page/body' %}</p>")
                assert html == u'<h1>Djedi</h1><p><span data-i18n="sv-se@page/body">Lightning fast!</span></p>'

        html = self.render(u"{% node 'foo/bar' default='bogus' %}")
        assert html == u'<span data-i18n="sv-se@foo/bar">bogus</span>'
        html = self.render(u"{% node 'l10n://foo/bar' default='bogus' %}")
        self.assertEqual(html, u'<span data-i18n="djedi@foo/bar">bogus</span>')

    def test_node_tag_with_default_scheme(self):
        cio.set('i18n://sv-se@page/title.txt', u'Swedish Djedi')
        html = self.render(u"{% node 'page/title' edit=False %}")
        assert html == u'Swedish Djedi'

        with self.settings(DJEDI={'URI_DEFAULT_SCHEME': 'l10n'}):
            html = self.render(u"{% node 'page/title' edit=False %}")
            assert html == u''

            cio.set('l10n://djedi@page/title.txt', u'Local Djedi')
            html = self.render(u"{% node 'page/title' edit=False %}")
            assert html == u'Local Djedi'

    def test_blocknode_tag(self):
        with self.assertRaises(TemplateSyntaxError):
            self.render("{% blocknode 'page/body' arg %}{% endblocknode %}")

        html = self.render(u"""
            {% blocknode 'page/body.md' edit=False %}
                # Djedi
                Lightning *fast*!
            {% endblocknode %}
        """)
        self.assertRenderedMarkdown(html, u'# Djedi\nLightning *fast*!')

        cio.set('i18n://sv-se@page/body.txt', u'Lightning fast!')
        html = self.render(u"""
            {% blocknode "page/body" %}
                Lorem ipsum
            {% endblocknode %}
        """)
        assert html == u'<span data-i18n="sv-se@page/body">Lightning fast!</span>'

        cio.set('i18n://sv-se@page/body.txt', u'')
        html = self.render(u"{% blocknode 'page/body' edit=False %}Lorem ipsum{% endblocknode %}")
        assert html == u''

    def test_blocknode_with_context(self):
        cio.set('i18n://sv-se@page/title.txt', u'Hej {name}!')

        source = u"""
            {% blocknode 'page/title' edit=False name=user.get_full_name %}
                Hello {name}!
            {% endblocknode %}
        """

        context = {'user': User(first_name=u'Jonas', last_name=u'Lundberg')}
        html = self.render(source, context)
        assert html == u'Hej Jonas Lundberg!'

        with cio.env(i18n='en-us'):
            html = self.render(source, context)
            assert html == u'Hello Jonas Lundberg!'

        html = self.render(u"""
            {% blocknode 'page/title' edit=False %}
                Hello {name}!
            {% endblocknode %}
        """)
        assert html == u'Hej {name}!'

    def test_collected_nodes(self):
        source = u"""
            {% node 'page/title' edit=False %}
            {% node 'page/title' default='fallback' edit=False %}
            {% node 'page/body' edit=False %}
        """
        pipeline.history.clear()
        self.render(source)
        assert len(pipeline.history) == 2

    def test_invalid_lazy_tag(self):
        with self.assertRaises(TemplateSyntaxError):
            register.lazy_tag('')

    def test_lazy_tag(self):
        @register.lazy_tag
        def foo():
            return lambda _: u'bar'
        html = self.render("{% foo %}")
        assert html == u'bar'

        @register.lazy_tag()
        def bar():
            return lambda _: u'foo'
        html = self.render("{% bar %}")
        assert html == u'foo'

    def test_djedi_admin_tag(self):
        source = u"""
            {% load djedi_admin %}
            {% djedi_admin %}
        """

        user = User(first_name=u'Jonas', last_name=u'Lundberg')

        class RequestMock(object):
            def __init__(self, user):
                self.user = user

        context = {'request': RequestMock(user=user)}
        html = self.render(source, context)
        assert html == u''

        user.is_superuser = True
        html = self.render(source, context)
        assert u'<script>window.DJEDI_NODES = {};</script>' in html



import django


def cmpt_context(context):
    if django.VERSION >= (1, 8):
        return context

    from django.template import Context
    return Context(context)


if django.VERSION < (1, 8):
    from django.template.loader import get_template_from_string
else:
    # Django 1.8+ has multiple template engines, we only test Django's for now.
    from django.template import engines

    def get_template_from_string(template_code):
        return engines['django'].from_string(template_code)


__all__ = ['cmpt_context', 'get_template_from_string']



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Node'
        db.create_table('djedi_node', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('key', self.gf('django.db.models.fields.CharField')(max_length=255, db_index=True)),
            ('content', self.gf('django.db.models.fields.TextField')(blank=True)),
            ('plugin', self.gf('django.db.models.fields.CharField')(max_length=8)),
            ('version', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('is_published', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('meta', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('date_created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
        ))
        db.send_create_signal(u'djedi', ['Node'])


    def backwards(self, orm):
        # Deleting model 'Node'
        db.delete_table('djedi_node')


    models = {
        u'djedi.node': {
            'Meta': {'object_name': 'Node'},
            'content': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'date_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_published': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'key': ('django.db.models.fields.CharField', [], {'max_length': '255', 'db_index': 'True'}),
            'meta': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'plugin': ('django.db.models.fields.CharField', [], {'max_length': '8'}),
            'version': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        }
    }

    complete_apps = ['djedi']


import logging

import cio

from django.core.exceptions import ImproperlyConfigured
from django.utils import translation

from cio.conf import settings
from cio.pipeline import pipeline
from djedi.auth import has_permission
from djedi.compat import reverse, NoReverseMatch
from djedi.utils.templates import render_embed

_log = logging.getLogger(__name__)


class TranslationMixin(object):

    def activate_language(self):
        # Activate current django translation
        language = translation.get_language()
        cio.env.push_state(i18n=language)


class AdminPanelMixin(object):

    def inject_admin_panel(self, request, response):
        # Do not inject admin panel on gzipped responses
        if 'gzip' in response.get('Content-Encoding', ''):
            _log.debug('gzip detected, not injecting panel.')
            return

        # Only inject admin panel in html pages
        content_type = response.get('Content-Type', '').split(';')[0]
        if content_type not in ('text/html', 'application/xhtml+xml'):
            _log.debug('Non-HTML Content-Type detected, not injecting')
            return

        # Do not inject admin panel in admin
        try:
            admin_prefix = reverse('admin:index')
        except NoReverseMatch:
            _log.debug(
                'No reverse match for "admin:index", can\'t detect '
                'django-admin pages'
            )
        else:
            if request.path.startswith(admin_prefix):
                _log.debug(
                    'admin page detected, not injecting panel. admin_prefix=%r',
                    admin_prefix
                )
                return

        try:
            djedi_cms_url = reverse('admin:djedi:cms')
        except NoReverseMatch:
            raise ImproperlyConfigured('Could not find djedi in your url conf, '
                                       'enable django admin or include '
                                       'djedi.urls within the admin namespace.')
        else:
            if request.path.startswith(djedi_cms_url):
                _log.debug('djedi page detected, not injecting panel')
                return

        # Validate user permissions
        if not has_permission(request):
            _log.debug('insufficient permissions, not injecting.')
            return

        embed = self.render_cms()
        self.body_append(response, embed)

    def render_cms(self):
        def get_requested_uri(node):
            # Get first namespace URI, remove any version and ensures extension.
            # TODO: Default extension fallback should be handled in content-io
            uri = node.namespace_uri
            uri = uri.clone(
                ext=uri.ext or settings.URI_DEFAULT_EXT,
                version=None,
            )
            return uri

        defaults = dict(
            (get_requested_uri(node), node.initial)
            for node in pipeline.history.list('get')
        )

        return render_embed(nodes=defaults)

    def body_append(self, response, html):
        idx = response.content.lower().rfind(b'</body>')

        if idx >= 0:
            response.content = b''.join((response.content[:idx],
                                         html.encode('utf8'),
                                         response.content[idx:]))

            if response.get('Content-Length', None):
                response['Content-Length'] = len(response.content)



import django

from .compat import include, patterns, url

app_name = 'djedi'

if django.VERSION < (2, 0):
    urlpatterns = patterns(
        url(r'^', include('djedi.admin.urls', namespace='djedi')),
    )
else:
    from django.urls import path
    urlpatterns = patterns(
        path('', include('djedi.admin.urls', namespace='djedi')),
    )



import cio
import six
import textwrap
from django import template
from django.template import TemplateSyntaxError
from .template import register
from ..compat import parse_bits


def render_node(node, context=None, edit=True):
    """
    Render node as html for templates, with edit tagging.
    """
    output = node.render(**context or {}) or u''
    if edit:
        return u'<span data-i18n="{0}">{1}</span>'.format(node.uri.clone(scheme=None, ext=None, version=None), output)
    else:
        return output


@register.lazy_tag
def node(key, default=None, edit=True):
    """
    Simple node tag:
    {% node 'page/title' default='Lorem ipsum' edit=True %}
    """
    node = cio.get(key, default=default or u'')
    return lambda _: render_node(node, edit=edit)


class BlockNode(template.Node):
    """
    Block node tag using body content as default:
    {% blocknode 'page/title' edit=True %}
        Lorem ipsum
    {% endblocknode %}
    """
    @classmethod
    def tag(cls, parser, token):
        # Parse tag args and kwargs
        bits = token.split_contents()[1:]
        params = ('uri', 'edit')
        args, kwargs = parse_bits(parser, bits, params, None, True, (True,), None, 'blocknode')

        # Assert uri is the only tag arg
        if len(args) > 1:
            raise TemplateSyntaxError('Malformed arguments to blocknode tag')

        # Resolve uri variable
        uri = args[0].resolve({})

        # Parse tag body (default content)
        tokens = parser.parse(('endblocknode',))
        parser.delete_first_token()  # Remove endblocknode tag

        # Render default content tokens and dedent common leading whitespace
        default = u''.join((token.render({}) for token in tokens))
        default = default.strip('\n\r')
        default = textwrap.dedent(default)

        # Get node for uri, lacks context variable lookup due to lazy loading.
        node = cio.get(uri, default)

        return cls(tokens, node, kwargs)

    def __init__(self, tokens, node, kwargs):
        self.tokens = tokens
        self.node = node
        self.kwargs = kwargs

    def render(self, context):
        # Resolve tag kwargs against context
        resolved_kwargs = dict((key, value.resolve(context)) for key, value in six.iteritems(self.kwargs))
        edit = resolved_kwargs.pop('edit', True)

        return render_node(self.node, context=resolved_kwargs, edit=edit)


register.tag('blocknode', BlockNode.tag)



import simplejson as json
import six
from django.http import HttpResponse
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

import cio
import cio.conf

from ..admin.mixins import JSONResponseMixin
from ..auth import has_permission
from ..utils.templates import render_embed


class APIView(JSONResponseMixin, View):

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super(APIView, self).dispatch(request, *args, **kwargs)


class EmbedApi(View):

    def get(self, request):
        if has_permission(request):
            return render_embed(request=request)
        else:
            # We used to `raise PermissionDenied` here (which might seem more
            # appropriate), but that has the annoying side effect of being
            # logged as an error in the browser dev tools, making people think
            # something is wrong.
            return HttpResponse(status=204)


class NodesApi(APIView):
    """
    JSON Response:
        {uri: content, uri: content, ...}
    """
    @never_cache
    def post(self, request):
        # Disable caching gets in CachePipe, defaults through this api is not trusted
        cio.conf.settings.configure(
            local=True,
            CACHE={
                'PIPE': {
                    'CACHE_ON_GET': False
                }
            }
        )

        nodes = []
        for uri, default in six.iteritems(json.loads(request.body)):
            node = cio.get(uri, default=default)
            nodes.append(node)

        data = dict((node.uri, node.content) for node in nodes)

        return self.render_to_json(data)



from cio.conf import settings
from cio.plugins import plugins
from djedi.tests import DjediTest


class SettingsTest(DjediTest):

    def test_settings(self):
        self.assertEqual(settings.THEME, 'luke')

    def test_plugin_settings(self):
        plugin = plugins.get('img')
        self.assertIn('foo', plugin.settings.keys())

    def test_default_scheme(self):
        self.assertEqual(settings.URI_DEFAULT_SCHEME, 'i18n')
        with self.settings(DJEDI={'URI_DEFAULT_SCHEME': 'l10n'}):
            self.assertEqual(settings.URI_DEFAULT_SCHEME, 'l10n')
        self.assertEqual(settings.URI_DEFAULT_SCHEME, 'i18n')



import json

from django.core.exceptions import ImproperlyConfigured

from ..compat import NoReverseMatch, render, render_to_string, reverse


def render_embed(nodes=None, request=None):
    context = {}

    if nodes is None:
        try:
            prefix = request.build_absolute_uri("/").rstrip("/")
            context.update(
                {
                    "cms_url": prefix + reverse("admin:djedi:cms"),
                    "exclude_json_nodes": True,
                }
            )
            output = render(request, "djedi/cms/embed.html", context)
        except NoReverseMatch:
            raise ImproperlyConfigured(
                "Could not find djedi in your url conf, "
                "enable django admin or include "
                "djedi.urls within the admin namespace."
            )

    else:
        context.update(
            {
                "cms_url": reverse("admin:djedi:cms"),
                "exclude_json_nodes": False,
                "json_nodes": json.dumps(nodes).replace("</", "\\x3C/"),
            }
        )
        output = render_to_string("djedi/cms/embed.html", context)

    return output



import cio
from cio.pipeline import pipeline


class DjediMiddleware(object):

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = self.process_request(request)
        if not response:
            try:
                response = self.get_response(request)
                response = self.process_response(request=request,
                                                 response=response)
            except Exception as e:
                self.process_exception(request=request, exception=e)
                raise
        return response

    def process_request(self, request):
        # Bootstrap content-io
        pipeline.clear()
        cio.env.reset()

    def process_response(self, request, response):
        pipeline.clear()
        return response

    def process_exception(self, request, exception):
        pipeline.clear()



import logging

_log = logging.getLogger(__name__)


def has_permission(request):
    user = getattr(request, 'user', None)
    if user:
        if user.is_superuser:
            return True

        if user.is_staff and user.groups.filter(name__iexact='djedi').exists():
            return True
    else:
        _log.warning("Request does not have `user` attribute. Make sure that "
                     "Djedi middleware is used after AuthenticationMiddleware")

    return False


def get_username(request):
    user = request.user
    if hasattr(user, 'get_username'):
        return user.get_username()
    else:
        return user.username



import six
from django.core.cache import InvalidCacheBackendError
from djedi.utils.encoding import smart_str, smart_unicode
from cio.backends.base import CacheBackend
from django.core.cache.backends.locmem import LocMemCache

from djedi.compat import get_cache


class DjangoCacheBackend(CacheBackend):

    def __init__(self, **config):
        """
        Get cache backend. Look for djedi specific cache first, then fallback on default
        """
        super(DjangoCacheBackend, self).__init__(**config)

        try:
            cache_name = self.config.get('NAME', 'djedi')
            cache = get_cache(cache_name)
        except (InvalidCacheBackendError, ValueError):
            from django.core.cache import cache

        self._cache = cache

    def clear(self):
        self._cache.clear()

    def _get(self, key):
        return self._cache.get(key)

    def _get_many(self, keys):
        return self._cache.get_many(keys)

    def _set(self, key, value):
        self._cache.set(key, value, timeout=None)  # TODO: Fix eternal timeout like viewlet

    def _set_many(self, data):
        self._cache.set_many(data, timeout=None)  # TODO: Fix eternal timeout like viewlet

    def _delete(self, key):
        self._cache.delete(key)

    def _delete_many(self, keys):
        self._cache.delete_many(keys)

    def _encode_content(self, uri, content):
        """
        Join node uri and content as string and convert to bytes to ensure no pickling in memcached.
        """
        if content is None:
            content = self.NONE
        return smart_str('|'.join([six.text_type(uri), content]))

    def _decode_content(self, content):
        """
        Split node string to uri and content and convert back to unicode.
        """
        content = smart_unicode(content)
        uri, _, content = content.partition(u'|')
        if content == self.NONE:
            content = None
        return uri or None, content


class DebugLocMemCache(LocMemCache):

    def __init__(self, *args, **kwargs):
        self.calls = 0
        self.hits = 0
        self.misses = 0
        self.sets = 0
        super(DebugLocMemCache, self).__init__(*args, **kwargs)

    def get(self, key, default=None, version=None, **kwargs):
        result = super(DebugLocMemCache, self).get(key, default=default, version=version)
        if kwargs.get('count', True):
            self.calls += 1
            if result is None:
                self.misses += 1
            else:
                self.hits += 1
        return result

    def get_many(self, keys, version=None):
        d = {}
        for k in keys:
            val = self.get(k, version=version, count=False)
            if val is not None:
                d[k] = val
        hits = len(d)
        self.calls += 1
        self.hits += hits
        self.misses += (len(keys) - hits)
        return d

    def set(self, *args, **kwargs):
        super(DebugLocMemCache, self).set(*args, **kwargs)
        self.calls += 1
        self.sets += 1

    def set_many(self, data, *args, **kwargs):
        result = super(DebugLocMemCache, self).set_many(data, *args, **kwargs)
        self.calls -= len(data)  # Remove calls from set()
        self.calls += 1
        return result



import django

from functools import partial

from django.shortcuts import render
from django.template.loader import render_to_string
from django.template.response import TemplateResponse as BaseTemplateResponse


if django.VERSION < (1, 6):
    from django.conf.urls.defaults import include, url
else:
    from django.conf.urls import include, url


if django.VERSION < (2, 0):
    from django.core.urlresolvers import reverse, NoReverseMatch
else:
    from django.urls import reverse, NoReverseMatch


def patterns(*urls):
    if django.VERSION < (1, 6):
        from django.conf.urls.defaults import patterns
        return patterns('', *urls)
    elif django.VERSION < (1, 10):
        from django.conf.urls import patterns
        return patterns('', *urls)
    return list(urls)


if django.VERSION >= (1, 9):
    if django.VERSION < (2, 0):
        from django.template.library import parse_bits
    else:
        def parse_bits(parser, bits, params, varargs, varkw, defaults,
                       takes_context, name):
            from django.template import library
            return library.parse_bits(
                parser=parser, bits=bits, params=params, varargs=varargs,
                varkw=varkw, defaults=defaults, kwonly=(), kwonly_defaults=(),
                takes_context=takes_context, name=name)

    def generic_tag_compiler(parser, token, params, varargs, varkw, defaults,
                             name, takes_context, node_class):
        """
        Returns a template.Node subclass.

        This got inlined into django.template.library since Django since 1.9, this here
        is a copypasta replacement:
        https://github.com/django/django/blob/stable/1.8.x/django/template/base.py#L1089
        """
        bits = token.split_contents()[1:]
        args, kwargs = parse_bits(parser, bits, params, varargs, varkw,
                                  defaults, takes_context, name)
        return node_class(takes_context, args, kwargs)
else:
    from django.template.base import parse_bits
    from django.template.base import generic_tag_compiler  # noqa


if django.VERSION >= (1, 8):
    # Always use the Django template engine on Django 1.8.
    render_to_string = partial(render_to_string, using='django')
    render = partial(render, using='django')

    class TemplateResponse(BaseTemplateResponse):
        def __init__(self, *args, **kwargs):
            kwargs['using'] = 'django'
            super(TemplateResponse, self).__init__(*args, **kwargs)
else:
    TemplateResponse = BaseTemplateResponse


if django.VERSION < (1, 9):
    from django.core.cache import get_cache
else:
    from django.core.cache import caches

    def get_cache(name):
        return caches[name]


__all__ = ['render_to_string',
           'render',
           'patterns',
           'include',
           'url',
           'reverse',
           'NoReverseMatch',
           'generic_tag_compiler',
           'parse_bits',
           'get_cache']
